from autoarray.plot.mat_wrap.wrap import wrap_2d


class LightProfileCentresScatter(wrap_2d.GridScatter):
    pass


class MassProfileCentresScatter(wrap_2d.GridScatter):
    pass


class MultipleImagesScatter(wrap_2d.GridScatter):
    pass


class CriticalCurvesPlot(wrap_2d.GridPlot):
    pass


class CausticsPlot(wrap_2d.GridPlot):
    pass
